import logging
import os
import zipfile
from typing import Dict, List, Optional

import yaml
from ibm_watson_machine_learning.client import APIClient
from mlflow.exceptions import ENDPOINT_NOT_FOUND, MlflowException

LOGGER = logging.getLogger(__name__)


def list_artifacts(client: APIClient) -> List[Dict]:
    """lists artifacts in WML repository

    Parameters
    ----------
    client : APIClient
        WML client

    Returns
    -------
    List[Dict]
        list of artifact details dictionary
    """
    artifacts = client.repository.get_details()["resources"]

    for artifact in artifacts:
        artifact["name"] = artifact["metadata"]["name"]

    return artifacts


def get_artifact_id_from_artifact_name(client: APIClient, artifact_name: str) -> str:
    """Returns artifact ID from artifact name

    Parameters
    ----------
    artifact_name : str
        artifact name

    Returns
    -------
    str
        artifact id
    """
    artifacts = list_artifacts(client=client)

    try:
        return next(item for item in artifacts if item["name"] == artifact_name)[
            "metadata"
        ]["id"]
    except StopIteration as _:
        message = f"artifact {artifact_name} not found"
        LOGGER.exception(message)
        raise MlflowException(message=message, error_code=ENDPOINT_NOT_FOUND)


def artifact_exists(client: APIClient, name: str) -> bool:
    """Checks if a artifact by the given name exists

    Parameters
    ----------
    client : APIClient
        WML client
    name : str
        name of the artifact

    Returns
    -------
    bool
        True if the artifact exists else False
    """
    artifacts = list_artifacts(client=client)

    return any(item for item in artifacts if item["name"] == name)


def list_deployments(client: APIClient) -> List[Dict]:
    """lists WML deployments

    Parameters
    ----------
    client : APIClient
        WML client

    Returns
    -------
    List[Dict]
        list of deployment details dictionary
    """
    deployments = client.deployments.get_details(get_all=True)["resources"]

    # `name` is a required key in each deployment
    for deployment in deployments:
        deployment["name"] = deployment["metadata"]["name"]

    return deployments


def get_deployment(client: APIClient, name: str) -> Dict:
    """retreive deployment details

    Parameters
    ----------
    client : APIClient
        WML client
    name : str
        name of the deployment

    Returns
    -------
    Dict
        deployment details dictionary
    """
    deployments = list_deployments(client=client)

    try:
        return next(item for item in deployments if item["name"] == name)

    except StopIteration as _:
        message = f"no deployment by the name {name} exists"
        LOGGER.exception(message)
        raise MlflowException(
            message=message,
            error_code=ENDPOINT_NOT_FOUND,
        )


def get_deployment_id_from_deployment_name(
    client: APIClient, deployment_name: str
) -> str:
    """Returns deployment ID from deployment name

    Parameters
    ----------
    deployment_name : str
        deployment name

    Returns
    -------
    str
        deployment id
    """
    return client.deployments.get_id(
        get_deployment(client=client, name=deployment_name)
    )


def deployment_exists(client: APIClient, name: str) -> bool:
    """Checks if a deployment by the given name exists

    Parameters
    ----------
    client : APIClient
        WML client
    name : str
        name of the deployment

    Returns
    -------
    bool
        True if the deployment exists else False
    """
    deployments = list_deployments(client=client)
    return any(item for item in deployments if item["name"] == name)


def get_space_id_from_space_name(client: APIClient, space_name: str) -> Optional[str]:
    """Returns space ID from the space name

    Parameters
    ----------
    client : APIClient
        WML client
    space_name : str
        space name

    Returns
    -------
    str | None
        space id
    """
    spaces = client.spaces.get_details(get_all=True)["resources"]

    try:
        return next(item for item in spaces if item["entity"]["name"] == space_name)[
            "metadata"
        ]["id"]
    except StopIteration as _:
        message = f"space {space_name} not found"
        LOGGER.warn(message)
        # raise MlflowException(message=message, error_code=ENDPOINT_NOT_FOUND)
        return None


def software_spec_exists(client: APIClient, name: str) -> bool:
    """Check if a given software specification exists.

    Parameters
    ----------
    client : APIClient
        WML client
    name : str
        name of the software specification

    Returns
    -------
    bool
        True if the software specification exists, else False
    """
    software_spec = client.software_specifications.get_id_by_name(sw_spec_name=name)
    return software_spec != "Not Found"


def get_software_spec_from_deployment_name(
    client: APIClient, deployment_name: str
) -> str:
    """Get software specification id for the given deployment

    Parameters
    ----------
    client : APIClient
        WML client
    deployment_name : str
        name of the deployment

    Returns
    -------
    str
        software specification id
    """
    deployment = get_deployment(client=client, name=deployment_name)
    artifact_id = deployment["entity"]["asset"]["id"]
    software_spec_id = client.repository.get_details(artifact_uid=artifact_id)[
        "entity"
    ]["software_spec"]["id"]

    return software_spec_id


def is_zipfile(file_path: str) -> bool:
    """Utility method to check if the given file path is a valid zip file.

    Parameters
    ----------
    file_path : str
        path to zip file

    Returns
    -------
    bool
        True if it is a valid zip file, else False
    """
    if not os.path.isfile(file_path):
        return False

    try:
        with zipfile.ZipFile(file_path) as zf:
            return zf.testzip() is None
    except zipfile.BadZipFile:
        return False


def refine_conda_yaml(conda_yaml: str) -> str:
    with open(conda_yaml, "r", encoding="utf-8") as f:
        env_data = yaml.safe_load(f)

    pip_dependencies = []

    for dep in env_data["dependencies"]:
        if isinstance(dep, dict):
            if "pip" in dep.keys():
                pip_dependencies.extend(dep["pip"])

    refined_env = {
        "channels": ["defaults"],
        "dependencies": [
            "pip",
            {"pip": pip_dependencies},
        ],
        "name": "mlflow-env",
    }

    with open(conda_yaml, "w", encoding="utf-8") as f:
        yaml.safe_dump(refined_env, f)


# TODO: implement logic to make sure the environment variables are set
def get_mlflow_config() -> Dict:
    return {
        # "MLFLOW_TRACKING_URI": os.environ.get("MLFLOW_TRACKING_URI", ""),
        "MLFLOW_S3_ENDPOINT_URL": os.environ.get("MLFLOW_S3_ENDPOINT_URL"),
        "AWS_SECRET_ACCESS_KEY": os.environ.get("AWS_SECRET_ACCESS_KEY"),
        "AWS_ACCESS_KEY_ID": os.environ.get("AWS_ACCESS_KEY_ID"),
    }



import json
import os

import pytest
from mlflow import MlflowException

from mlflow_watsonml.config import Config


@pytest.fixture(scope="session")
def create_correct_credentials(tmp_path_factory: pytest.TempPathFactory):
    credentials = {
        "username": "username",
        "apikey": "apikey",
        "url": "url",
        "instance_id": "openshift",
        "version": "4.0",
    }

    credentials_path = tmp_path_factory.mktemp("credentials") / "wml_credentials.json"
    with open(credentials_path, "w") as f:
        json.dump(credentials, f)

    return credentials_path


def test_config_input_arg_pass():
    input = dict()
    input["apikey"] = "apikey"
    input["location"] = "location"
    input["url"] = "url"

    config = Config(config=input)

    assert config["wml_credentials"] == {"apikey": "apikey", "url": "url"}


def test_config_input_arg_fail():
    input = dict()
    input["apikey"] = "apikey"

    with pytest.raises(MlflowException):
        _ = Config(config=input)


def test_config_input_env_var_pass(create_correct_credentials):
    with pytest.MonkeyPatch.context() as mp:
        mp.setenv("WML_CREDENTIALS_FILE", str(create_correct_credentials))

        config = Config()

        assert config["wml_credentials"] == {
            "username": "username",
            "apikey": "apikey",
            "url": "url",
            "instance_id": "openshift",
            "version": "4.0",
        }


def test_config_input_env_var_fail():
    with pytest.MonkeyPatch.context() as mp:
        mp.delenv("WML_CREDENTIALS_FILE", raising=False)
        mp.delenv("URL", raising=False)
        mp.setenv("APIKEY", "apikey")

        with pytest.raises(MlflowException):
            _ = Config()



import logging
from typing import Any, Dict, Optional, Tuple

from ibm_watson_machine_learning.client import APIClient
from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import NOT_IMPLEMENTED

from mlflow_watsonml.store import *
from mlflow_watsonml.utils import *

LOGGER = logging.getLogger(__name__)


def deploy(
    client: APIClient,
    name: str,
    artifact_id: str,
    revision_id: str,
    batch: bool = False,
    environment_variables: Optional[Dict] = None,
    hardware_spec_id: Optional[str] = None,
) -> Dict:
    """Create a new WML deployment

    Parameters
    ----------
    client : APIClient
        WML client
    name : str
        name of the new deployment to create
    artifact_id : str
        UID of the model or function stored in WML repository
    batch : bool, optional
        whether to use batch or online method of deployment,
        by default False

    Returns
    -------
    Dict
        deployment details dictionary
    """
    if environment_variables is None:
        environment_variables = dict()

    if batch:
        deployment_props = {
            client.deployments.ConfigurationMetaNames.NAME: name,
            client.deployments.ConfigurationMetaNames.BATCH: {},
            client.deployments.ConfigurationMetaNames.CUSTOM: environment_variables,
            client.deployments.ConfigurationMetaNames.HARDWARE_SPEC: {
                "id": hardware_spec_id
            }
            if hardware_spec_id is not None
            else {},
            client.deployments.ConfigurationMetaNames.ASSET: {
                "id": artifact_id,
                "rev": revision_id,
            },
        }
    else:
        deployment_props = {
            client.deployments.ConfigurationMetaNames.NAME: name,
            client.deployments.ConfigurationMetaNames.CUSTOM: environment_variables,
            client.deployments.ConfigurationMetaNames.HARDWARE_SPEC: {
                "id": hardware_spec_id
            }
            if hardware_spec_id is not None
            else {},
            client.deployments.ConfigurationMetaNames.ONLINE: {},
            client.deployments.ConfigurationMetaNames.ASSET: {
                "id": artifact_id,
                "rev": revision_id,
            },
        }

    try:
        deployment_details = client.deployments.create(
            artifact_uid=artifact_id,
            meta_props=deployment_props,
        )

        deployment_details["name"] = deployment_details["metadata"]["name"]

        LOGGER.info(deployment_details)
        LOGGER.info(f"Created {'batch' if batch else 'online'} deployment - {name}")

    except Exception as e:
        raise MlflowException(e)

    return deployment_details


def delete_deployment(client: APIClient, name: str):
    """Delete an existing deployment from WML.
    This method deletes the deployment and all the artifacts associated with it.

    Parameters
    ----------
    client : APIClient
        WML client
    name : str
        name of the deployment to delete
    """
    try:
        deployment_details = get_deployment(client=client, name=name)

        deployment_id = client.deployments.get_id(deployment_details=deployment_details)
        client.deployments.delete(deployment_uid=deployment_id)
        LOGGER.info(f"Deleted deployment {name} with id {deployment_id}.")

        artifact_id = deployment_details["entity"]["asset"]["id"]
        artifact_details = client.repository.get_details(artifact_uid=artifact_id)
        client.repository.delete(artifact_uid=artifact_id)
        LOGGER.info(
            f"Deleted artifact {name} with id {artifact_id} from the repository."
        )

        software_spec_id = artifact_details["entity"]["software_spec"]["id"]
        software_spec_details = client.software_specifications.get_details(
            sw_spec_uid=software_spec_id
        )
        software_spec_name = software_spec_details["metadata"]["name"]

        if software_spec_name == f"{name}_sw_spec":
            client.software_specifications.delete(sw_spec_uid=software_spec_id)
            LOGGER.info(
                f"Deleted software specification {software_spec_name} with id {software_spec_id} from the repository."
            )

    except Exception as e:
        LOGGER.exception(e)
        raise MlflowException(e)


def update_deployment(
    client: APIClient,
    name: str,
    artifact_id: str,
    revision_id: str,
) -> Dict:
    deployment_id = get_deployment_id_from_deployment_name(
        client=client, deployment_name=name
    )
    metadata = {
        client.deployments.ConfigurationMetaNames.ASSET: {
            "id": artifact_id,
            "rev": revision_id,
        }
    }

    updated_deployment = client.deployments.update(
        deployment_uid=deployment_id, changes=metadata
    )

    LOGGER.info(updated_deployment)

    return updated_deployment


def store_or_update_artifact(
    client: APIClient,
    model_uri: str,
    artifact_name: str,
    flavor: str,
    software_spec_id: str,
    artifact_id: Optional[str] = None,
    environment_variables: Optional[Dict] = None,
) -> Tuple[str, str]:
    if flavor == "sklearn":
        artifact_id, revision_id = store_sklearn_artifact(
            client=client,
            model_uri=model_uri,
            artifact_name=artifact_name,
            software_spec_id=software_spec_id,
            artifact_id=artifact_id,
        )

    elif flavor == "onnx":
        artifact_id, revision_id = store_onnx_artifact(
            client=client,
            model_uri=model_uri,
            artifact_name=artifact_name,
            software_spec_id=software_spec_id,
            artifact_id=artifact_id,
        )

    elif flavor == "watson_nlp":
        artifact_id, revision_id = store_watson_nlp_artifact(
            client=client,
            model_uri=model_uri,
            artifact_name=artifact_name,
            software_spec_id=software_spec_id,
            artifact_id=artifact_id,
            config=environment_variables,
        )

    else:
        raise MlflowException(
            f"Flavor {flavor} is invalid or not implemented",
            error_code=NOT_IMPLEMENTED,
        )

    return (artifact_id, revision_id)


def create_custom_software_spec(
    client: APIClient,
    name: str,
    custom_packages: Optional[List[str]],
    conda_yaml: Optional[str] = None,
    rewrite: bool = False,
) -> str:
    if software_spec_exists(client=client, name=name):
        if rewrite:
            while software_spec_exists(client=client, name=name):
                software_spec_id = client.software_specifications.get_id_by_name(
                    sw_spec_name=name
                )
                client.software_specifications.delete(sw_spec_uid=software_spec_id)

                LOGGER.info(
                    f"Deleted software specification {name} with id {software_spec_id}"
                )
        else:
            LOGGER.warn(
                f"""Software spec {name} already exists."""
                """skipping software spec creation."""
                """restart with rewrite=True if software spec needs to be updated."""
            )

            software_spec_id = client.software_specifications.get_id_by_name(name)
            return software_spec_id

    try:
        base_software_spec_id = client.software_specifications.get_id_by_name(
            "runtime-22.2-py3.10"
        )

        meta_prop_sw_spec = {
            client.software_specifications.ConfigurationMetaNames.NAME: name,
            client.software_specifications.ConfigurationMetaNames.BASE_SOFTWARE_SPECIFICATION: {
                "guid": base_software_spec_id
            },
        }

        sw_spec_details = client.software_specifications.store(
            meta_props=meta_prop_sw_spec
        )
        software_spec_id = client.software_specifications.get_id(sw_spec_details)

        if conda_yaml is not None:
            if not os.path.exists(conda_yaml):
                raise FileNotFoundError(f"conda.yaml file not found!")

            refine_conda_yaml(conda_yaml=conda_yaml)

            pkg_extn_name = f"{name}_conda_env"
            # pkg_extn_id =

            # if client.package_extensions.get_id_by_name(pkg_extn_name=pkg_extn_name) != "Not Found":

            meta_prop_pkg_extn = {
                client.package_extensions.ConfigurationMetaNames.NAME: pkg_extn_name,
                client.package_extensions.ConfigurationMetaNames.TYPE: "conda_yml",
            }

            pkg_extn_details = client.package_extensions.store(
                meta_props=meta_prop_pkg_extn,
                file_path=conda_yaml,
            )

            pkg_extn_id = client.package_extensions.get_uid(pkg_extn_details)

            client.software_specifications.add_package_extension(
                software_spec_id, pkg_extn_id
            )

        if custom_packages is not None:
            for custom_package in custom_packages:
                if not is_zipfile(custom_package):
                    raise MlflowException(f"{custom_package} is not a valid zip file.")

                pkg_name = os.path.splitext(os.path.basename(custom_package))[0]

                meta_prop_pkg_extn = {
                    client.package_extensions.ConfigurationMetaNames.NAME: pkg_name,
                    client.package_extensions.ConfigurationMetaNames.TYPE: "pip_zip",
                }

                pkg_extn_details = client.package_extensions.store(
                    meta_props=meta_prop_pkg_extn, file_path=custom_package
                )

                pkg_extn_id = client.package_extensions.get_id(pkg_extn_details)

                client.software_specifications.add_package_extension(
                    software_spec_id, pkg_extn_id
                )

    except Exception as e:
        LOGGER.exception(e)

        raise MlflowException(e)

    LOGGER.info(
        f"Successfully created {name} software specification with ID {software_spec_id}"
    )

    return software_spec_id



import zipfile

import pytest
from mlflow import MlflowException
from pytest import LogCaptureFixture, MonkeyPatch
from resources.mock.mock_client import MockAPIClient

import mlflow_watsonml.deploy
from mlflow_watsonml.deploy import WatsonMLDeploymentClient
from mlflow_watsonml.utils import *

MOCK_WML_CREDENTIALS = {
    "username": "user",
    "apikey": "correct_api_key",
    "url": "https://url",
    "instance_id": "wml",
    "version": "1.0",
}


@pytest.fixture(autouse=True)
def mock_client(monkeypatch: MonkeyPatch):
    # Mock the APIClient
    monkeypatch.setattr(mlflow_watsonml.deploy, "APIClient", MockAPIClient)


def test_list_artifacts():
    client = WatsonMLDeploymentClient(config=MOCK_WML_CREDENTIALS).get_wml_client(
        endpoint="space_1"
    )

    artifacts = list_artifacts(client=client)

    assert isinstance(artifacts, list)
    assert len(artifacts) == 3

    for artifact in artifacts:
        assert isinstance(artifact, dict)
        assert "name" in artifact.keys()


def test_get_artifact_id_from_artifact_name_success():
    client = WatsonMLDeploymentClient(config=MOCK_WML_CREDENTIALS).get_wml_client(
        endpoint="space_1"
    )

    artifact_id = get_artifact_id_from_artifact_name(
        client=client, artifact_name="artifact_1"
    )

    assert artifact_id == "id_of_artifact_1"


def test_get_artifact_id_from_artifact_name_exception(caplog: LogCaptureFixture):
    client = WatsonMLDeploymentClient(config=MOCK_WML_CREDENTIALS).get_wml_client(
        endpoint="space_1"
    )

    with pytest.raises(MlflowException):
        artifact_id = get_artifact_id_from_artifact_name(
            client=client, artifact_name="artifact_01"
        )

    assert "artifact artifact_01 not found" in caplog.text


def test_artifact_exists():
    client = WatsonMLDeploymentClient(config=MOCK_WML_CREDENTIALS).get_wml_client(
        endpoint="space_1"
    )

    assert artifact_exists(client=client, name="artifact_1")
    assert not artifact_exists(client=client, name="artifact_01")


def test_list_deployments():
    client = WatsonMLDeploymentClient(config=MOCK_WML_CREDENTIALS).get_wml_client(
        endpoint="space_1"
    )

    deployments = list_deployments(client=client)

    assert isinstance(deployments, list)
    assert len(deployments) == 2

    for deployment in deployments:
        assert isinstance(deployment, dict)
        assert "name" in deployment.keys()


def test_get_deployment_success():
    client = WatsonMLDeploymentClient(config=MOCK_WML_CREDENTIALS).get_wml_client(
        endpoint="space_1"
    )

    deployment = get_deployment(client=client, name="deployment_1")

    assert isinstance(deployment, dict)
    assert "name" in deployment.keys()
    assert "metadata" in deployment.keys()
    assert isinstance(deployment["metadata"], dict)
    assert "id" in deployment["metadata"].keys()


def test_get_deployment_exception(caplog: LogCaptureFixture):
    client = WatsonMLDeploymentClient(config=MOCK_WML_CREDENTIALS).get_wml_client(
        endpoint="space_1"
    )

    with pytest.raises(MlflowException):
        deployment = get_deployment(client=client, name="deployment_01")

    assert "no deployment by the name deployment_01 exists" in caplog.text


def test_get_deployment_id_from_deployment_name_success():
    client = WatsonMLDeploymentClient(config=MOCK_WML_CREDENTIALS).get_wml_client(
        endpoint="space_1"
    )

    deployment_id = get_deployment_id_from_deployment_name(
        client=client, deployment_name="deployment_1"
    )

    assert deployment_id == "id_of_deployment_1"


def test_get_deployment_id_from_deployment_name_exception(caplog: LogCaptureFixture):
    client = WatsonMLDeploymentClient(config=MOCK_WML_CREDENTIALS).get_wml_client(
        endpoint="space_1"
    )

    with pytest.raises(MlflowException):
        deployment_id = get_deployment_id_from_deployment_name(
            client=client, deployment_name="deployment_01"
        )

    assert "no deployment by the name deployment_01 exists" in caplog.text


def test_deployment_exists():
    client = WatsonMLDeploymentClient(config=MOCK_WML_CREDENTIALS).get_wml_client(
        endpoint="space_1"
    )

    assert deployment_exists(client=client, name="deployment_1")
    assert not deployment_exists(client=client, name="deployment_01")


def test_get_space_id_from_space_name_success():
    client = mlflow_watsonml.deploy.APIClient(wml_credentials=MOCK_WML_CREDENTIALS)

    space_id = get_space_id_from_space_name(client=client, space_name="space_1")

    assert space_id == "id_of_space_1"


@pytest.fixture
def zip_file_path(tmp_path):
    # Create a temporary zip file for testing
    file_path = tmp_path / "test.zip"
    with zipfile.ZipFile(file_path, "w") as zf:
        zf.writestr("file1.txt", "Hello, World!")
        zf.writestr("file2.txt", "artifact Factory")

    yield str(file_path)

    # Clean up the temporary file
    os.remove(file_path)


def test_is_zipfile_valid(zip_file_path):
    assert is_zipfile(zip_file_path) is True


def test_is_zipfile_invalid(zip_file_path):
    # Modify the file extension to make it invalid
    invalid_file_path = zip_file_path.replace(".zip", ".txt")
    assert is_zipfile(invalid_file_path) is False


def test_is_zipfile_nonexistent():
    non_existent_file_path = "/path/to/nonexistent.zip"
    assert is_zipfile(non_existent_file_path) is False


def test_is_zipfile_corrupt_zip(zip_file_path):
    # Overwrite the zip file with a corrupt file
    with open(zip_file_path, "w") as f:
        f.write("corrupt file data")

    assert is_zipfile(zip_file_path) is False


def test_get_software_spec_from_deployment_name_success():
    client = WatsonMLDeploymentClient(config=MOCK_WML_CREDENTIALS).get_wml_client(
        endpoint="space_1"
    )

    sw_spec = get_software_spec_from_deployment_name(
        client=client, deployment_name="deployment_1"
    )

    assert sw_spec == "id_of_sw_spec_1"


def test_get_software_spec_from_deployment_name_exception(caplog: LogCaptureFixture):
    client = WatsonMLDeploymentClient(config=MOCK_WML_CREDENTIALS).get_wml_client(
        endpoint="space_1"
    )

    with pytest.raises(MlflowException):
        sw_spec = get_software_spec_from_deployment_name(
            client=client, deployment_name="deployment_3"
        )

    assert f"no deployment by the name deployment_3 exists"



import logging
import os
from types import FunctionType
from typing import Any, Dict, Optional, Tuple

import mlflow
from ibm_watson_machine_learning.client import APIClient
from mlflow.exceptions import MlflowException

LOGGER = logging.getLogger(__name__)


def store_or_update_model(
    client: APIClient,
    model_object: Any,
    model_name: str,
    model_type: str,
    software_spec_uid: str,
    model_id: Optional[str] = None,
) -> Tuple[str, str]:
    """Store or update a model object in a WML repository

    Parameters
    ----------
    client : APIClient
        WML client
    model_object : Any
        model object
    model_name : str
        name of the model
    model_type : str
        type of model
    software_spec_uid : str
        UID of software specification
    model_id : str, optional
        asset id of the model to be updated
        by default None

    Returns
    -------
    Tuple[str, str]
        model id, model revision id
    """

    try:
        if model_id is None:
            model_props = {
                client.repository.ModelMetaNames.NAME: model_name,
                client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: software_spec_uid,
                client.repository.ModelMetaNames.TYPE: model_type,
            }
            model_details = client.repository.store_model(
                model=model_object,
                meta_props=model_props,
                training_data=None,
                training_target=None,
                feature_names=None,
                label_column_names=None,
            )
            LOGGER.info(model_details)
            LOGGER.info(f"Stored model {model_name} in the repository.")

            model_id = client.repository.get_model_id(model_details=model_details)
        else:
            model_props = {
                client.repository.ModelMetaNames.NAME: model_name,
                client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: software_spec_uid,
                client.repository.ModelMetaNames.TYPE: model_type,
            }
            model_details = client.repository.update_model(
                model_uid=model_id,
                updated_meta_props=model_props,
                update_model=model_object,
            )
            LOGGER.info(model_details)
            LOGGER.info(f"Updated model {model_name} in the repository.")

            model_id = client.repository.get_model_id(model_details=model_details)

        revision_details = client.repository.create_model_revision(model_uid=model_id)

        rev_id = revision_details["metadata"]["rev"]
        LOGGER.info(revision_details)
        LOGGER.info(
            f"Created model revision for model {model_name} and version {rev_id}"
        )

    except Exception as e:
        raise MlflowException(e)

    return (model_id, rev_id)


def store_or_update_function(
    client: APIClient,
    deployable_function: FunctionType,
    function_name: str,
    software_spec_uid: str,
    function_id: Optional[str] = None,
) -> Tuple[str, str]:
    """Store or update a python function in WML repository

    Parameters
    ----------
    client : APIClient
        WML client
    deployable_function : FunctionType
        function to deploy
    function_name : str
        name of the function
    software_spec_uid : str
        software specification id
    function_id: str, optional
        asset id of the function to be updated
        by default None

    Returns
    -------
    Tuple[str, str]
        function id and function revisin id
    """
    try:
        if function_id is None:
            metaprops = {
                client.repository.FunctionMetaNames.NAME: function_name,
                client.repository.FunctionMetaNames.SOFTWARE_SPEC_ID: software_spec_uid,
            }
            function_details = client.repository.store_function(
                function=deployable_function,
                meta_props=metaprops,
            )
            LOGGER.info(function_details)
            LOGGER.info(f"Stored function {function_name} in the repository.")

            function_id = client.repository.get_function_id(
                function_details=function_details
            )
        else:
            metaprops = {
                client.repository.FunctionMetaNames.NAME: function_name,
            }
            function_details = client.repository.update_function(
                function_uid=function_id,
                changes=metaprops,
                update_function=deployable_function,
            )
            LOGGER.info(function_details)
            LOGGER.info(f"Updated function {function_name} in the repository.")

            function_id = client.repository.get_function_id(
                function_details=function_details
            )

        revision_details = client.repository.create_function_revision(
            function_uid=function_id
        )
        rev_id = revision_details["metadata"]["rev"]
        LOGGER.info(revision_details)
        LOGGER.info(
            f"Created function revision for function {function_name} and version {rev_id}"
        )
    except Exception as e:
        LOGGER.exception(e)
        raise MlflowException(e)

    return (function_id, rev_id)


def store_onnx_artifact(
    client: APIClient,
    model_uri: str,
    artifact_name: str,
    software_spec_id: str,
    artifact_id: Optional[str] = None,
) -> Tuple[str, str]:
    """store onnx artifact in WML

    Parameters
    ----------
    client : APIClient
        WML client
    model_uri : str
        model URI
    artifact_name : str
        name of the artifact
    software_spec_id : str
        id of software specification
    artifact_id : Optional[str], optional
        artifact id of the stored model, by default None

    Returns
    -------
    Tuple[str, str]
        model id, revision id
    """

    # the args have to be passed as default value in the scorer
    def deployable_onnx_scorer(artifact_uri=model_uri):
        import os
        import tempfile

        import mlflow
        import onnx  # type: ignore
        from onnxruntime import InferenceSession  # type: ignore

        def score(payload: dict):
            artifact_dir = os.path.join(tempfile.gettempdir(), "artifacts")

            # `download_artifacts` returns the local path if it's already been downloaded
            artifact_file = mlflow.artifacts.download_artifacts(
                artifact_uri=artifact_uri, dst_path=artifact_dir
            )
            model_file = os.path.join(artifact_file, "model.onnx")
            onnx.checker.check_model(model_file)  # type: ignore
            model = onnx.load(model_file)
            input_name = model.graph.input[0].name

            scoring_output = {"predictions": []}

            sess = InferenceSession(model.SerializeToString())

            for data in payload["input_data"]:
                values = data.get("values")
                # fields = data.get("fields")
                predictions = sess.run(None, {input_name: values})[0]

                scoring_output["predictions"].append({"values": predictions.tolist()})

            return scoring_output

        return score

    function_id, rev_id = store_or_update_function(
        client=client,
        deployable_function=deployable_onnx_scorer,
        function_name=artifact_name,
        software_spec_uid=software_spec_id,
        function_id=artifact_id,
    )

    return (function_id, rev_id)


def store_sklearn_artifact(
    client: APIClient,
    model_uri: str,
    artifact_name: str,
    software_spec_id: str,
    artifact_id: Optional[str] = None,
) -> Tuple[str, str]:
    """store sklearn artifact in WML

    Parameters
    ----------
    client : APIClient
        WML client
    model_uri : str
        model URI
    artifact_name : str
        name of the artifact
    software_spec_id : str
        id of software specification
    artifact_id : Optional[str], optional
        artifact id of the stored model, by default None

    Returns
    -------
    Tuple[str, str]
        model id, revision id
    """
    model_object = mlflow.sklearn.load_model(model_uri=model_uri)
    model_id, rev_id = store_or_update_model(
        client=client,
        model_object=model_object,
        model_name=artifact_name,
        model_type="scikit-learn_1.1",
        software_spec_uid=software_spec_id,
        model_id=artifact_id,
    )

    return (model_id, rev_id)


def store_watson_nlp_artifact(
    client: APIClient,
    model_uri: str,
    artifact_name: str,
    software_spec_id: str,
    artifact_id: Optional[str] = None,
    config: Optional[Dict] = None,
) -> Tuple[str, str]:
    """store watson nlp artifact in WML

    Parameters
    ----------
    client : APIClient
        WML client
    model_uri : str
        model URI
    artifact_name : str
        name of the artifact
    software_spec_id : str
        id of software specification
    artifact_id : Optional[str], optional
        artifact id of the stored model, by default None

    Returns
    -------
    Tuple[str, str]
        model id, revision id
    """

    # the args have to be passed as default value in the scorer
    def deployable_watson_nlp_scorer(artifact_uri=model_uri, config=config):
        import os
        import tempfile

        import mlflow
        import watson_nlp  # type: ignore

        for key, val in config.items():  # type: ignore
            os.environ[key] = val

        def score(payload: dict):
            artifact_dir = os.path.join(tempfile.gettempdir(), "artifacts")

            # `download_artifacts` returns the local path if it's already been downloaded
            artifact_file = mlflow.artifacts.download_artifacts(
                artifact_uri=artifact_uri, dst_path=artifact_dir
            )

            model = watson_nlp.load(artifact_file)

            scoring_output = {"predictions": []}

            for data in payload["input_data"]:
                values = data.get("values")
                # fields = data.get("fields")
                predictions = model.run_batch(values)
                predictions = [prediction.to_dict() for prediction in predictions]

                scoring_output["predictions"].append({"values": predictions})

            return scoring_output

        return score

    function_id, rev_id = store_or_update_function(
        client=client,
        deployable_function=deployable_watson_nlp_scorer,
        function_name=artifact_name,
        software_spec_uid=software_spec_id,
        function_id=artifact_id,
    )

    return (function_id, rev_id)



import json
import os
from typing import Dict, Optional

from dotenv import load_dotenv
from mlflow import MlflowException
from mlflow.protos.databricks_pb2 import INVALID_PARAMETER_VALUE

from mlflow_watsonml.logging import LOGGER

WML_CREDENTIALS = "wml_credentials"
WML_CREDENTIALS_FILE = "wml_credentials_file"
APIKEY = "apikey"
LOCATION = "location"
URL = "url"
TOKEN = "token"
USERNAME = "username"
PASSWORD = "password"
INSTANCE_ID = "instance_id"
VERSION = "version"


class Config(dict):
    def __init__(self, config: Optional[Dict[str, str]] = None):
        """Initializes constants from input `config`, `.env` file or environment variables

        Parameters
        ----------
        config : Optional[Dict[str, str]], optional
            wml credentials, by default None

        Raises
        ------
        MlflowException
            missing credentials
        """
        super().__init__()

        if config is None:
            LOGGER.info(
                "Input credentials not provided. Attempting to load credentials from .env file"
            )

            ret = load_dotenv()

            if ret:
                LOGGER.info("Loaded input credentials from .env file")
            else:
                LOGGER.info(
                    ".env file not found. Attempting to read credentials from environment variables."
                )

            config = {}

            if "WML_CREDENTIALS_FILE" in os.environ.keys():
                config[WML_CREDENTIALS_FILE] = os.getenv("WML_CREDENTIALS_FILE")

            if "APIKEY" in os.environ.keys():
                config[APIKEY] = os.getenv("APIKEY")

            if "LOCATION" in os.environ.keys():
                config[LOCATION] = os.getenv("LOCATION")

            if "URL" in os.environ.keys():
                config[URL] = os.getenv("URL")

            if "TOKEN" in os.environ.keys():
                config[TOKEN] = os.getenv("TOKEN")

            if "USERNAME" in os.environ.keys():
                config[USERNAME] = os.getenv("USERNAME")

            if "PASSWORD" in os.environ.keys():
                config[PASSWORD] = os.getenv("PASSWORD")

            if "INSTANCE_ID" in os.environ.keys():
                config[INSTANCE_ID] = os.getenv("INSTANCE_ID")

            if "VERSION" in os.environ.keys():
                config[VERSION] = os.getenv("VERSION")

        # Load the appropriate environment variables based on their presence
        if WML_CREDENTIALS_FILE in config:
            wml_credentials_file = config[WML_CREDENTIALS_FILE]

            with open(wml_credentials_file, "r") as f:
                self[WML_CREDENTIALS] = json.load(f)

        elif APIKEY in config and LOCATION in config and URL in config:
            apikey = config[APIKEY]
            location = config[LOCATION]
            url = config[URL]

            self[WML_CREDENTIALS] = {
                APIKEY: apikey,
                URL: url,
            }

        elif TOKEN in config and LOCATION in config and URL in config:
            token = config[TOKEN]
            location = config[LOCATION]
            url = config[URL]

            self[WML_CREDENTIALS] = {
                TOKEN: token,
                URL: url,
            }

        elif (
            USERNAME in config
            and PASSWORD in config
            and URL in config
            and INSTANCE_ID in config
            and VERSION in config
        ):
            username = config[USERNAME]
            password = config[PASSWORD]
            url = config[URL]
            instance_id = config[INSTANCE_ID]
            version = config[VERSION]

            self[WML_CREDENTIALS] = {
                USERNAME: username,
                PASSWORD: password,
                URL: url,
                INSTANCE_ID: instance_id,
                VERSION: version,
            }

        elif (
            USERNAME in config
            and APIKEY in config
            and URL in config
            and INSTANCE_ID in config
            and VERSION in config
        ):
            username = config[USERNAME]
            apikey = config[APIKEY]
            url = config[URL]
            instance_id = config[INSTANCE_ID]
            version = config[VERSION]

            self[WML_CREDENTIALS] = {
                USERNAME: username,
                APIKEY: apikey,
                URL: url,
                INSTANCE_ID: instance_id,
                VERSION: version,
            }

        elif (
            TOKEN in config
            and URL in config
            and INSTANCE_ID in config
            and VERSION in config
        ):
            token = config[TOKEN]
            url = config[URL]
            instance_id = config[INSTANCE_ID]
            version = config[VERSION]

            self[WML_CREDENTIALS] = {
                TOKEN: token,
                URL: url,
                INSTANCE_ID: instance_id,
                VERSION: version,
            }

        else:
            raise MlflowException(
                "Missing Credentials", error_code=INVALID_PARAMETER_VALUE
            )



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()


with open("requirements.txt", "r") as fh:
    install_requires = fh.readlines()

exec(open("mlflow_watsonml/_version.py").read())

setup(
    name="mlflow-watsonml",
    version=__version__,  # type: ignore
    description="WatsonML MLflow deployment plugin",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/IBM/mlflow-watsonml",
    packages=["mlflow_watsonml"],
    author="IBM AI Model Factory team",
    author_email="dhruv.shah@ibm.com",
    install_requires=install_requires,
    extras_require={
        "dev": ["ipython", "black", "pytest", "build", "wheel", "twine", "pytest-cov"],
        "onnx": ["onnx", "onnxruntime"],
        "docs": ["mkdocs", "mkdocstrings-python", "mkdocs-material"],
    },
    entry_points={"mlflow.deployments": "watsonml=mlflow_watsonml.deploy"},
    python_requires=">=3.9",
    # use_scm_version=True,
    # setup_requires=["setuptools_scm"],
)



import pytest
from mlflow import MlflowException
from pytest import LogCaptureFixture, MonkeyPatch
from resources.mock.mock_client import MockAPIClient

import mlflow_watsonml.deploy
from mlflow_watsonml.deploy import WatsonMLDeploymentClient

MOCK_WML_CREDENTIALS = {
    "username": "user",
    "apikey": "correct_api_key",
    "url": "https://url",
    "instance_id": "wml",
    "version": "1.0",
}


@pytest.fixture(autouse=True)
def mock_client(monkeypatch: MonkeyPatch):
    # Mock the APIClient
    monkeypatch.setattr(mlflow_watsonml.deploy, "APIClient", MockAPIClient)


def test_connect_success():
    client = WatsonMLDeploymentClient(config=MOCK_WML_CREDENTIALS)

    # Assert that the _wml_client attribute is set
    assert isinstance(client._wml_client, MockAPIClient)


def test_connect_exception(caplog: LogCaptureFixture):
    # Call the connect method with mock credentials
    mock_wml_credentials = MOCK_WML_CREDENTIALS.copy()
    mock_wml_credentials["apikey"] = "incorrect_api_key"

    assert MOCK_WML_CREDENTIALS["apikey"] != "incorrect_api_key"

    with pytest.raises(Exception):
        _ = WatsonMLDeploymentClient(config=mock_wml_credentials)

    # Assert that the exception was logged
    assert "Connection Failed!" in caplog.text


def test_get_wml_client_success():
    client = WatsonMLDeploymentClient(config=MOCK_WML_CREDENTIALS)

    wml_client = client.get_wml_client(endpoint="space_1")

    assert isinstance(wml_client, MockAPIClient)


def test_get_wml_client_exception(caplog: LogCaptureFixture):
    client = WatsonMLDeploymentClient(config=MOCK_WML_CREDENTIALS)

    with pytest.raises(MlflowException):
        _ = client.get_wml_client(endpoint="space_3")

    assert "space space_3 not found" in caplog.text


def test_create_deployment_success(monkeypatch: MonkeyPatch):
    ...


def test_create_deployment_exception(caplog: LogCaptureFixture):
    ...


def test_update_deployment_success():
    ...


def test_update_deployment_exception(caplog: LogCaptureFixture):
    ...


def test_delete_deployment_success():
    ...


def test_delete_deployment_exception(caplog: LogCaptureFixture):
    ...


def test_list_deployments_success():
    ...


def test_list_deployments_exception(caplog: LogCaptureFixture):
    ...


def test_get_deployment_success():
    ...


def test_get_deployment_exception(caplog: LogCaptureFixture):
    ...


def test_predict_success():
    ...


def test_predict_exception(caplog: LogCaptureFixture):
    ...



from typing import Dict, List, Optional, Union

import mlflow
import numpy as np
import pandas as pd
from ibm_watson_machine_learning.client import APIClient
from mlflow.deployments import BaseDeploymentClient
from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import ENDPOINT_NOT_FOUND, INVALID_PARAMETER_VALUE

from mlflow_watsonml.config import Config
from mlflow_watsonml.logging import LOGGER
from mlflow_watsonml.utils import *
from mlflow_watsonml.wml import *


def target_help():
    # TODO: Improve
    help_string = (
        "The mlflow-watsonml plugin integrates IBM WatsonML "
        "with the MLFlow deployments API.\n\n"
        "Before using this plugin, you must set up a json file "
        "containing WML credentials and create an environment variable "
        "along with deployment space name. "
    )
    return help_string


def run_local(name, model_uri, flavor=None, config=None):
    raise MlflowException("mlflow-watsonml does not currently support run_local.")


class WatsonMLDeploymentClient(BaseDeploymentClient):
    def __init__(self, target_uri: str = "watsonml", config: Optional[Dict] = None):
        """Initialize a WML `APIClient`. The method has an optional parameter called
        `config` which should have the WML credentials. If `config` is `None`, then
        the plugin will try to search for WML credentials in `.env` file or the
        environment variables.

        Refer to the following links for setting up the credentials -

        1. [Cloud Pak for Data as a Service](https://ibm.github.io/watson-machine-learning-sdk/setup_cloud.html#authentication)
        2. [Cloud Pak for Data](https://ibm.github.io/watson-machine-learning-sdk/setup_cpd.html#authentication)

        Parameters
        ----------
        target_uri : str, optional
            Target URI for mlflow deployment, by default "watsonml"
        config : Optional[Dict], optional
            WML Credentials, by default None
        """
        super().__init__(target_uri)

        self.wml_config = Config(config=config)
        self.connect(wml_credentials=self.wml_config["wml_credentials"])

    def connect(self, wml_credentials: Dict) -> None:
        """Connect to WML APIClient and set the default deployment space

        Parameters
        ----------
        wml_credentials : Dict
            WML Credentials
        """
        try:
            client = APIClient(wml_credentials=wml_credentials)
            LOGGER.info("Connected to WML Client successfully")

        except Exception as e:
            LOGGER.exception(e)
            raise MlflowException(
                "Could not establish connection, check WML credentials." f"{e}",
                error_code=ENDPOINT_NOT_FOUND,
            )

        self._wml_client = client

    def get_wml_client(self, endpoint: str) -> APIClient:
        """Returns WML API client

        Parameters
        ----------
        endpoint : str
            deployment space name

        Returns
        -------
        APIClient
            WML client
        """
        client = self._wml_client

        try:
            space_uid = get_space_id_from_space_name(
                client=client,
                space_name=endpoint,
            )

            if space_uid is None:
                raise MlflowException(
                    f"Endpoint {endpoint} not found.",
                    error_code=ENDPOINT_NOT_FOUND,
                )
            client.set.default_space(space_uid=space_uid)

            LOGGER.info(
                f"Set deployment space to {endpoint} with space id - {space_uid}"
            )

        except Exception as e:
            LOGGER.exception(e)
            raise MlflowException(f"Failed to set deployment space {endpoint}", f"{e}")

        return client

    def create_deployment(
        self,
        name: str,
        model_uri: str,
        flavor: str,
        config: Optional[Dict],
        endpoint: str,
    ) -> Dict:
        """Deploy a model at `model_uri` to a WML target. this method blocks until
        deployment completes (i.e. until it's possible to perform inference with the deployment).
        In the case of conflicts (e.g. if it's not possible to create the specified deployment
        without due to conflict with an existing deployment), raises a
        :py:class:`mlflow.exceptions.MlflowException`.

        Parameters
        ----------
        name : str
            name of the deployment
        model_uri : str
            URI (local or remote) of the model
        flavor : str
            flavor of the deployed model
        config : Dict
            configuration parameters for wml deployment.
            possible optional configuration keys are -
            - "software_spec_name" : name of the software specification to reuse
            - "conda_yaml" : filepath of conda.yaml file
            - "custom_packages": a list of str - zip file paths of the packages
            - "rewrite_software_spec": bool whether to rewrite the software spec
            - "hardware_spec_name" : name of the hardware specification to use (Default: XS)
        endpoint : str
            deployment space name

        Returns
        -------
        Dict
            deployment details dictionary
        """
        client = self.get_wml_client(endpoint=endpoint)

        if config is None:
            config = dict()

        # check if a deployment by that name exists
        if deployment_exists(client=client, name=name):
            raise MlflowException(
                f"Deployment {name} already exists. Use `update_deployment()` or use a different name",
                error_code=INVALID_PARAMETER_VALUE,
            )

        if "software_spec_name" in config.keys():
            software_spec_id = client.software_specifications.get_id_by_name(
                config["software_spec_name"]
            )

            if software_spec_id == "Not Found":
                raise MlflowException(
                    f"Software Specification {config['software_spec_name']} not found.",
                    error_code=INVALID_PARAMETER_VALUE,
                )

        else:
            if "conda_yaml" in config.keys():
                conda_yaml = config["conda_yaml"]
            else:
                conda_yaml = mlflow.pyfunc.get_model_dependencies(
                    model_uri=model_uri, format="conda"
                )  # other option is to have a default conda_yaml for each flavor

            custom_packages: List[str] = config.get("custom_packages")
            rewrite: bool = config.get("rewrite_software_spec", False)

            software_spec_id = create_custom_software_spec(
                client=client,
                name=f"{name}_sw_spec",
                custom_packages=custom_packages,
                conda_yaml=conda_yaml,
                rewrite=rewrite,
            )

        artifact_name = f"{name}_v1"
        environment_variables = get_mlflow_config()

        artifact_id, revision_id = store_or_update_artifact(
            client=client,
            model_uri=model_uri,
            artifact_name=artifact_name,
            flavor=flavor,
            software_spec_id=software_spec_id,
            environment_variables=environment_variables,
        )

        batch = config.get("batch", False)

        hardware_spec_name = config.get("hardware_spec_name", "XS")
        if hardware_spec_name is not None:
            hardware_spec_id = client.hardware_specifications.get_id_by_name(
                hardware_spec_name
            )
            if hardware_spec_id == "Not Found":
                LOGGER.warn(
                    f"Hardware Specification - {hardware_spec_name} not found. Using default."
                )
                hardware_spec_id = None
        else:
            hardware_spec_id = None

        deployment_details = deploy(
            client=client,
            name=name,
            artifact_id=artifact_id,
            revision_id=revision_id,
            batch=batch,
            environment_variables=environment_variables,
            hardware_spec_id=hardware_spec_id,
        )

        return deployment_details

    def update_deployment(
        self,
        name: str,
        model_uri: str,
        flavor: str,
        config: Optional[Dict],
        endpoint: str,
    ) -> Dict:
        """Update the deployment with the specified name. You can update the URI of the model, the
        flavor of the deployed model (in which case the model URI must also be specified). By default,
        this method blocks until deployment completes (i.e. until it's possible to perform inference
        with the updated deployment).

        Parameters
        ----------
        name : str
            Unique name of the deployment to update
        model_uri : str
            URI of a new model to deploy
        flavor : str
            new model flavor to use for deployment. If provided,
            ``model_uri`` must also be specified.
        config : Optional[Dict], optional
            dict containing updated WML-specific configuration for the
        endpoint : str
            deployment space name

        Returns
        -------
        Dict
            deployment details dictionary
        """
        client = self.get_wml_client(endpoint=endpoint)

        if config is None:
            config = dict()

        # check if a deployment by that name exists
        if not deployment_exists(client=client, name=name):
            raise MlflowException(
                f"Deployment {name} doesn't exist. Use `create_deployment()`",
                error_code=INVALID_PARAMETER_VALUE,
            )

        current_deployment = self.get_deployment(name=name, endpoint=endpoint)
        artifact_id = current_deployment["entity"]["asset"]["id"]
        artifact_rev = int(current_deployment["entity"]["asset"]["rev"])

        new_artifact_name = f"{name}_v{artifact_rev+1}"

        if "software_spec_name" in config.keys():
            software_spec_id = client.software_specifications.get_id_by_name(
                config["software_spec_name"]
            )

            if software_spec_id == "Not Found":
                raise MlflowException(
                    f"Software Specification {config['software_spec_name']} not found.",
                    error_code=INVALID_PARAMETER_VALUE,
                )

        else:
            if "conda_yaml" in config.keys():
                conda_yaml = config["conda_yaml"]
            else:
                conda_yaml = mlflow.pyfunc.get_model_dependencies(
                    model_uri=model_uri, format="conda"
                )  # other option is to have a default conda_yaml for each flavor

            custom_packages: List[str] = config.get("custom_packages")

            software_spec_id = create_custom_software_spec(
                client=client,
                name=f"{name}_sw_spec",
                custom_packages=custom_packages,
                conda_yaml=conda_yaml,
                rewrite=True,
            )
        environment_variables = get_mlflow_config()
        artifact_id, revision_id = store_or_update_artifact(
            client=client,
            model_uri=model_uri,
            artifact_name=new_artifact_name,
            flavor=flavor,
            software_spec_id=software_spec_id,
            environment_variables=environment_variables,
        )

        deployment_details = update_deployment(
            client=client,
            name=name,
            artifact_id=artifact_id,
            revision_id=revision_id,
        )

        return deployment_details

    def delete_deployment(
        self, name: str, config: Optional[Dict] = None, endpoint: Optional[str] = None
    ):
        """Delete the deployment with name ``name`` from WML. Deletion is idempotent
        (i.e. deletion does not fail if retried on a non-existent deployment).

        Parameters
        ----------
        name : str
            name of the deployment to delete
        config : Optional[Dict], optional
            configuration parameters, by default None
        endpoint : Optional[str], optional
            deployment space name, by default None
        """
        if config is None:
            config = dict()

        client = self.get_wml_client(endpoint=endpoint)

        if deployment_exists(client=client, name=name):
            delete_deployment(client=client, name=name)

    def list_deployments(self, endpoint: str):
        """List deployments. This method returns an unpaginated list of all deployments

        Parameters
        ----------
        endpoint : str
            deployment space name

        Returns
        -------
        List[Dict]
            A list of dicts corresponding to deployments. Each dict is guaranteed to
            contain a 'name' key containing the deployment name. The other fields of
            the returned dictionary and their types follow WML deployment details convention.
        """
        return list_deployments(client=self.get_wml_client(endpoint=endpoint))

    def get_deployment(self, name: str, endpoint: str):
        """Returns a dictionary describing the specified deployment, throwing a
        :py:class:`mlflow.exceptions.MlflowException` if no deployment exists with the provided
        name.
        The dict is guaranteed to contain a 'name' key containing the deployment name.
        The other fields of the returned dictionary and their types follow WML convention.

        Parameters
        ----------
        name : str
            name of the deployment to fetch
        endpoint : str
            deployment space name

        Returns
        -------
        Dict
            A dict corresponding to the retrieved deployment. The dict is guaranteed to
            contain a 'name' key corresponding to the deployment name.
        """
        return get_deployment(client=self.get_wml_client(endpoint=endpoint), name=name)

    def predict(
        self,
        deployment_name: str,
        inputs: Union[pd.DataFrame, np.ndarray, List[Any], Dict[str, Any]],
        endpoint: str,
    ) -> Union[np.ndarray, pd.DataFrame, pd.Series, List]:
        """Compute predictions on inputs using the specified deployment

        predict(
            model_input:
        ) ->

        Parameters
        ----------
        deployment_name : str
            Name of deployment to predict against
        inputs : pd.DataFrame
            Input data (or arguments) to pass to the deployment for inference,
        endpoint : str
            deployment space name

        Returns
        -------
        pd.DataFrame
            Model predictions as pandas.DataFrame
        """
        client = self.get_wml_client(endpoint=endpoint)

        deployment_details = self.get_deployment(
            name=deployment_name, endpoint=endpoint
        )

        scoring_payload = {
            client.deployments.ScoringMetaNames.INPUT_DATA: [{"values": inputs}]
        }

        if "custom" in deployment_details["entity"].keys():
            scoring_payload[
                client.deployments.ScoringMetaNames.ENVIRONMENT_VARIABLES
            ] = deployment_details["entity"]["custom"]

        deployment_id = client.deployments.get_id(deployment_details=deployment_details)

        predictions = client.deployments.score(
            deployment_id=deployment_id, meta_props=scoring_payload
        )["predictions"]

        if (
            len(predictions) > 0
            and isinstance(predictions[0], dict)
            and "fields" in predictions[0].keys()
        ):
            fields = predictions[0]["fields"]

            frames = []
            for prediction in predictions:
                frames.extend(prediction["values"])

            ans = pd.DataFrame(frames, columns=fields)
            return ans

        return predictions  # ans

    def explain(self, deployment_name=None, df=None, endpoint=None):
        raise NotImplementedError()

    def create_endpoint(self, name: str, config: Optional[Dict] = None) -> Dict:
        """
        Create an endpoint with the specified target. By default, this method should block until
        creation completes (i.e. until it's possible to create a deployment within the endpoint).
        In the case of conflicts (e.g. if it's not possible to create the specified endpoint
        due to conflict with an existing endpoint), raises a
        :py:class:`mlflow.exceptions.MlflowException`. See target-specific plugin documentation
        for additional detail on support for asynchronous creation and other configuration.

        :param name: Unique name to use for endpoint. If another endpoint exists with the same
                     name, raises a :py:class:`mlflow.exceptions.MlflowException`.
        :param config: (optional) Dict containing target-specific configuration for the
                       endpoint.
        :return: Dict corresponding to created endpoint, which must contain the 'name' key.
        """
        client = self._wml_client

        if config is None:
            config = dict()

        metadata = dict()
        metadata[client.spaces.ConfigurationMetaNames.NAME] = name

        if client.spaces.ConfigurationMetaNames.DESCRIPTION in config.keys():
            metadata[client.spaces.ConfigurationMetaNames.DESCRIPTION] = config[
                client.spaces.ConfigurationMetaNames.DESCRIPTION
            ]

        endpoint_details = client.spaces.store(
            meta_props=metadata, background_mode=False
        )

        return endpoint_details

    def update_endpoint(self, endpoint, config=None):
        """
        Update the endpoint with the specified name. You can update any target-specific attributes
        of the endpoint (via `config`). By default, this method should block until the update
        completes (i.e. until it's possible to create a deployment within the endpoint). See
        target-specific plugin documentation for additional detail on support for asynchronous
        update and other configuration.

        :param endpoint: Unique name of endpoint to update
        :param config: (optional) dict containing target-specific configuration for the
                       endpoint
        :return: None
        """
        raise NotImplementedError()

    def delete_endpoint(self, endpoint):
        """
        Delete the endpoint from the specified target. Deletion should be idempotent (i.e. deletion
        should not fail if retried on a non-existent deployment).

        :param endpoint: Name of endpoint to delete
        :return: None
        """
        client = self._wml_client

        endpoint_id = get_space_id_from_space_name(client=client, space_name=endpoint)

        if endpoint_id is not None:
            client.spaces.delete(space_id=endpoint_id)

    def list_endpoints(self):
        """
        List endpoints in the specified target. This method is expected to return an
        unpaginated list of all endpoints (an alternative would be to return a dict with
        an 'endpoints' field containing the actual endpoints, with plugins able to specify
        other fields, e.g. a next_page_token field, in the returned dictionary for pagination,
        and to accept a `pagination_args` argument to this method for passing
        pagination-related args).

        :return: A list of dicts corresponding to endpoints. Each dict is guaranteed to
                 contain a 'name' key containing the endpoint name. The other fields of
                 the returned dictionary and their types may vary across targets.
        """
        client = self._wml_client
        endpoints = client.spaces.get_details(get_all=True)["resources"]

        for endpoint in endpoints:
            endpoint["name"] = endpoint["entity"]["name"]

        return endpoints

    def get_endpoint(self, endpoint):
        client = self._wml_client
        deployment_space_id = get_space_id_from_space_name(
            client=client, space_name=endpoint
        )
        endpoint_details = client.spaces.get_details(space_id=deployment_space_id)
        return endpoint_details

    def create_custom_wml_spec(
        self,
        name: str,
        custom_packages: Optional[List[str]],
        conda_yaml: Optional[str],
        endpoint: str,
        rewrite: bool = False,
    ) -> str:
        """Create a custom WML Software Specification

        Parameters
        ----------
        name : str
            name for the software specification
        custom_packages : Optional[List[str]]
            a list of zip file paths for custom packages
        conda_yaml : Optional[str]
            file path to conda.yaml file
        endpoint : str
            deployment space name
        rewrite : bool, optional
            whether to rewrite the existing software specification, by default False

        Returns
        -------
        str
            id of the software specification
        """
        client = self.get_wml_client(endpoint=endpoint)
        software_spec_id = create_custom_software_spec(
            client=client,
            name=name,
            custom_packages=custom_packages,
            conda_yaml=conda_yaml,
            rewrite=rewrite,
        )

        return software_spec_id
